import { PeerId, Libp2p } from "@libp2p/interface";
import { GossipSub } from "@chainsafe/libp2p-gossipsub";
import { AbstractGossipLog } from "./AbstractGossipLog.js";
import { SignedMessage } from "./SignedMessage.js";
import { ServiceMap } from "./interface.js";
export declare const getSyncProtocol: (topic: string) => string;
export declare const getPushProtocol: (topic: string) => string;
export interface GossipLogServiceInit {
    maxInboundStreams?: number;
    maxOutboundStreams?: number;
}
export declare class GossipLogService<Payload = unknown> {
    #private;
    readonly libp2p: Libp2p<ServiceMap>;
    readonly messageLog: AbstractGossipLog<Payload>;
    private static extractGossipSub;
    readonly syncProtocol: string;
    readonly pushProtocol: string;
    private readonly log;
    private readonly maxInboundStreams;
    private readonly maxOutboundStreams;
    private readonly syncQueue;
    private readonly syncQueuePeers;
    private readonly controller;
    private readonly pushTopology;
    private readonly litePeers;
    pubsub: GossipSub | null;
    constructor(libp2p: Libp2p<ServiceMap>, messageLog: AbstractGossipLog<Payload>, init: GossipLogServiceInit);
    start(): Promise<void>;
    private handleConnectionOpen;
    private handleConnectionClose;
    stop(): Promise<void>;
    publish(signedMessage: SignedMessage<Payload>, options?: {
        sourceId?: string;
    }): Promise<PeerId[]>;
    private handleMessage;
    private handleUpdate;
    private getPushSink;
    private handleIncomingPush;
    private handleIncomingSync;
    private scheduleSync;
    private sync;
    private retry;
    private wait;
}
