var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var _GossipLogService_instances, _GossipLogService_pushTopologyId, _GossipLogService_started, _GossipLogService_pushStreams, _GossipLogService_sync, _GossipLogService_push;
import { TopicValidatorResult, CodeError, } from "@libp2p/interface";
import { GossipSub, multicodec as GossipsubIDv11 } from "@chainsafe/libp2p-gossipsub";
import { logger } from "@libp2p/logger";
import { peerIdFromString } from "@libp2p/peer-id";
import * as lp from "it-length-prefixed";
import { pipe } from "it-pipe";
import { pushable } from "it-pushable";
import { equals } from "uint8arrays";
import { anySignal } from "any-signal";
import PQueue from "p-queue";
import { assert, DelayableController } from "@canvas-js/utils";
import { Event } from "#protocols/events";
import { DEFAULT_PROTOCOL_SELECT_TIMEOUT, MAX_INBOUND_STREAMS, MAX_OUTBOUND_STREAMS, NEGOTIATE_FULLY, SYNC_RETRY_INTERVAL, SYNC_RETRY_LIMIT, SYNC_TIMEOUT, } from "./constants.js";
import { decodeId, encodeId } from "./ids.js";
import { Client, Server, decodeRequests, encodeResponses } from "./sync/index.js";
import { MISSING_PARENT, SyncTimeoutError, wait } from "./utils.js";
export const getSyncProtocol = (topic) => `/canvas/v1/${topic}/sync`;
export const getPushProtocol = (topic) => `/canvas/v1/${topic}/push`;
export class GossipLogService {
    static extractGossipSub(libp2p) {
        try {
            const pubsub = libp2p.services.pubsub;
            assert(pubsub instanceof GossipSub, "expected pubsub instanceof GossipSub");
            return pubsub;
        }
        catch (err) {
            return null;
        }
    }
    constructor(libp2p, messageLog, init) {
        _GossipLogService_instances.add(this);
        this.libp2p = libp2p;
        this.messageLog = messageLog;
        this.syncProtocol = getSyncProtocol(this.messageLog.topic);
        this.pushProtocol = getPushProtocol(this.messageLog.topic);
        this.syncQueue = new PQueue({ concurrency: 1 });
        this.syncQueuePeers = new Set();
        this.controller = new AbortController();
        this.pushTopology = new Set();
        this.litePeers = new Set();
        this.pubsub = null;
        _GossipLogService_pushTopologyId.set(this, null);
        _GossipLogService_started.set(this, false);
        _GossipLogService_pushStreams.set(this, new Map());
        this.handleConnectionOpen = ({ detail: connection }) => {
            this.log("connection:open %s %p", connection.id, connection.remotePeer);
        };
        this.handleConnectionClose = ({ detail: connection }) => {
            this.log("connection:close %s %p", connection.id, connection.remotePeer);
        };
        this.handleMessage = ({ detail: { msgId, propagationSource, msg } }) => {
            if (msg.topic !== this.messageLog.topic) {
                return;
            }
            this.log("received gossipsub message %s via %p", msgId, propagationSource);
            const sourceId = propagationSource.toString();
            let event;
            try {
                event = Event.decode(msg.data);
            }
            catch (err) {
                this.log.error("error decoding gossipsub message: %O", err);
                this.pubsub?.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
                return;
            }
            if (event.insert === undefined) {
                this.log("ignoring gossipsub message %s", msgId);
                this.pubsub?.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Ignore);
                return;
            }
            let signedMessage;
            try {
                signedMessage = this.messageLog.decode(event.insert.value);
                assert(equals(event.insert.key, signedMessage.key), "invalid key");
            }
            catch (err) {
                this.log.error("invalid message: %O", err);
                this.pubsub?.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
                return;
            }
            this.messageLog.insert(signedMessage, { peerId: sourceId, publish: false }).then(({ id }) => {
                this.pubsub?.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Accept);
                let count = 0;
                for (const peer of this.litePeers) {
                    __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_push).call(this, peer, msg.data);
                    count += 1;
                }
                if (count > 0) {
                    this.log("forwarded message %s to %d lite peers", id, count);
                }
            }, (err) => {
                if (err instanceof CodeError && err.code === MISSING_PARENT) {
                    this.pubsub?.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Ignore);
                    this.scheduleSync(propagationSource);
                }
                else {
                    this.log.error("rejecting gossipsub message %s: %O", msgId, err);
                    this.pubsub?.reportMessageValidationResult(msgId, sourceId, TopicValidatorResult.Reject);
                }
            });
        };
        this.getPushSink = (connection, stream) => async (msgs) => {
            const sourceId = connection.remotePeer.toString();
            try {
                for await (const msg of msgs) {
                    const event = Event.decode(msg.subarray());
                    if (event.insert !== undefined) {
                        const signedMessage = this.messageLog.decode(event.insert.value);
                        assert(equals(event.insert.key, signedMessage.key), "invalid key");
                        try {
                            await this.messageLog.insert(signedMessage, { publish: false, peerId: sourceId });
                        }
                        catch (err) {
                            if (err instanceof CodeError && err.code === MISSING_PARENT) {
                                this.scheduleSync(connection.remotePeer);
                                continue;
                            }
                            else {
                                throw err;
                            }
                        }
                        this.publish(signedMessage, { sourceId });
                    }
                    else if (event.update !== undefined) {
                        await this.handleUpdate(connection.remotePeer, event.update);
                    }
                }
            }
            catch (err) {
                stream.abort(err instanceof Error ? err : new Error(`${err}`));
            }
        };
        this.handleIncomingPush = ({ connection, stream }) => {
            const peerId = connection.remotePeer;
            this.log("opened incoming push stream %s from peer %p", stream.id, peerId);
            const source = pushable();
            __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").set(peerId.toString(), { stream, source });
            Promise.all([
                // outgoing events
                pipe(source, lp.encode, stream.sink),
                // incoming events
                pipe(stream.source, lp.decode, this.getPushSink(connection, stream)).then(() => {
                    source.end();
                    stream.close();
                    __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").delete(peerId.toString());
                }),
            ]).finally(() => {
                __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").delete(peerId.toString());
                this.log("closed incoming push stream %s from peer %p", stream.id, peerId);
            });
        };
        this.handleIncomingSync = async ({ connection, stream }) => {
            const peerId = connection.remotePeer;
            this.log("opened incoming sync stream %s from peer %p", stream.id, peerId);
            const timeoutController = new DelayableController(SYNC_TIMEOUT);
            const signal = anySignal([this.controller.signal, timeoutController.signal]);
            signal.addEventListener("abort", (err) => {
                if (stream.status === "open") {
                    stream.abort(new Error("TIMEOUT"));
                }
            });
            try {
                await this.messageLog.serve(async (txn) => {
                    const server = new Server(this.messageLog.topic, txn);
                    await pipe(stream.source, lp.decode, decodeRequests, async function* (reqs) {
                        for await (const req of reqs) {
                            timeoutController.delay();
                            yield req;
                        }
                    }, (reqs) => server.handle(reqs), encodeResponses, lp.encode, stream.sink);
                });
                this.log("closed incoming stream %s from peer %p", stream.id, peerId);
            }
            catch (err) {
                if (err instanceof Error && err.message === "TIMEOUT") {
                    this.log.error("timed out incoming stream %s from peer %p", stream.id, peerId);
                    stream.abort(err);
                }
                else if (err instanceof Error) {
                    this.log.error("aborting incoming stream %s from peer %p: %O", stream.id, peerId, err);
                    stream.abort(err);
                }
                else {
                    this.log.error("aborting incoming stream %s from peer %p: %O", stream.id, peerId, err);
                    stream.abort(new Error("internal error"));
                }
            }
            finally {
                timeoutController.clear();
                signal.clear();
            }
        };
        this.log = logger(`canvas:gossiplog:[${this.messageLog.topic}]:service`);
        this.pubsub = GossipLogService.extractGossipSub(libp2p);
        this.maxInboundStreams = init.maxInboundStreams ?? MAX_INBOUND_STREAMS;
        this.maxOutboundStreams = init.maxOutboundStreams ?? MAX_OUTBOUND_STREAMS;
    }
    async start() {
        this.log("start");
        if (this.libp2p.status !== "started")
            return;
        if (__classPrivateFieldGet(this, _GossipLogService_started, "f") === true)
            return;
        __classPrivateFieldSet(this, _GossipLogService_started, true, "f");
        this.libp2p.addEventListener("connection:open", this.handleConnectionOpen);
        this.libp2p.addEventListener("connection:close", this.handleConnectionClose);
        this.pubsub?.addEventListener("gossipsub:message", this.handleMessage);
        /**
         * sync is the protocol for initiating merkle syncs using the Okra index.
         * syncs are initiated in three cases:
         * 1. when a peer receives an orphan message via gossipsub, it schedules
         *    a sync with the gossipsub message's "propagation source"
         * 2. when a peer receives a push update with missing heads, it schedules
         *    a sync with the sender
         * 3. when a peer receives a push insert with missing parents, it schedules
         *    a sync with the sender
         */
        await this.libp2p.handle(this.syncProtocol, this.handleIncomingSync, {
            maxInboundStreams: this.maxInboundStreams,
            maxOutboundStreams: this.maxOutboundStreams,
        });
        /**
     * push is the protocol for sending Update and Insert events to specific peers.
     * - Update events carry the sender's set of latest concurrent message ids ("heads")
     * - Insert events carry a signed message as an encoded [key, value] entry
     * Updates are pushed in two cases:
     * 1. once for every new connection, so that the recipient can merkle sync
     *    back if necessary. this happens from both sides of each connection simultaneously.
     * 2. whenever a peer finishes an merkle sync during which it received one or more
     *    missing messages, it pushes its new heads to all of its peers, except for
     *    the sync source.
     * Inserts are used for interacting with "lite client" peers, identified by the lack of
     * support for the gossipsub protocol.
     * - lite clients always push every new append directly to all of their peers.
     * - full clients forward every gossipsub message to all of their lite client peers.
     * - full clients forward every insert push to all of their lite client peers, except
         for the push sender.
     */
        await this.libp2p.handle(this.pushProtocol, this.handleIncomingPush, {
            maxInboundStreams: this.maxInboundStreams,
            maxOutboundStreams: this.maxOutboundStreams,
        });
        __classPrivateFieldSet(this, _GossipLogService_pushTopologyId, await this.libp2p.register(this.pushProtocol, {
            notifyOnTransient: false,
            onConnect: (peerId, connection) => {
                this.log("connected to %p", peerId);
                this.pushTopology.add(peerId.toString());
                const protocolSelectSignal = AbortSignal.timeout(DEFAULT_PROTOCOL_SELECT_TIMEOUT);
                connection.newStream(this.pushProtocol, { negotiateFully: NEGOTIATE_FULLY, signal: protocolSelectSignal }).then((stream) => {
                    const peerId = connection.remotePeer;
                    this.log("opened outgoing push stream %s to peer %p", stream.id, peerId);
                    const source = pushable();
                    __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").set(peerId.toString(), { stream, source });
                    Promise.all([
                        // outgoing events
                        pipe(source, lp.encode, stream.sink),
                        // incoming events
                        pipe(stream.source, lp.decode, this.getPushSink(connection, stream)),
                    ]).finally(() => this.log("closed outgoing push stream %s from peer %p", stream.id, peerId));
                    this.messageLog.getClock().then(([_, heads]) => {
                        __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_push).call(this, peerId.toString(), Event.encode({ update: { heads: heads.map(encodeId) } }));
                    });
                }, (err) => {
                    this.log.error("failed to open outgoing push stream: %O", err);
                    throw err;
                });
                this.libp2p.peerStore.get(peerId).then((peer) => {
                    if (!peer.protocols.includes(GossipsubIDv11)) {
                        this.litePeers.add(peerId.toString());
                    }
                });
            },
            onDisconnect: (peerId) => {
                this.log("disconnected %p", peerId);
                this.pushTopology.delete(peerId.toString());
                this.litePeers.delete(peerId.toString());
                const { source, stream } = __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").get(peerId.toString()) ?? {};
                if (source !== undefined && stream !== undefined) {
                    __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").delete(peerId.toString());
                    source.end();
                    stream.close();
                }
            },
        }), "f");
        this.pubsub?.subscribe(this.messageLog.topic);
    }
    async stop() {
        this.log("stop");
        if (this.libp2p.status !== "started")
            return;
        if (__classPrivateFieldGet(this, _GossipLogService_started, "f") === false)
            return;
        __classPrivateFieldSet(this, _GossipLogService_started, false, "f");
        this.controller.abort();
        this.pubsub?.unsubscribe(this.messageLog.topic);
        for (const [peer, { stream, source }] of __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f")) {
            stream.close();
            source.end();
            __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").delete(peer);
        }
        await this.libp2p.unhandle(this.syncProtocol);
        await this.libp2p.unhandle(this.pushProtocol);
        if (__classPrivateFieldGet(this, _GossipLogService_pushTopologyId, "f") !== null) {
            this.libp2p.unregister(__classPrivateFieldGet(this, _GossipLogService_pushTopologyId, "f"));
            __classPrivateFieldSet(this, _GossipLogService_pushTopologyId, null, "f");
        }
        this.libp2p.removeEventListener("connection:open", this.handleConnectionOpen);
        this.libp2p.removeEventListener("connection:close", this.handleConnectionClose);
        this.pubsub?.removeEventListener("gossipsub:message", this.handleMessage);
    }
    async publish(signedMessage, options = {}) {
        if (!__classPrivateFieldGet(this, _GossipLogService_started, "f")) {
            return [];
        }
        const { id, key, value } = signedMessage;
        const event = { insert: { key, value } };
        const data = Event.encode(event);
        if (this.pubsub === null) {
            // If we're a lite client, then push directly to all of our topology peers
            const recipients = [];
            for (const peer of __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").keys()) {
                if (peer === options.sourceId)
                    continue;
                __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_push).call(this, peer, data); // TODO: doesn't guarantee we actually pushed to the peer...
                recipients.push(peerIdFromString(peer));
            }
            return recipients;
        }
        else {
            // If we're a full client, publish to pubsub, and still push directly to all of our lite topology peers
            const recipients = [];
            for (const peer of this.litePeers) {
                if (peer === options.sourceId)
                    continue;
                __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_push).call(this, peer, data);
                // TODO: doesn't guarantee we actually pushed to the peer
                recipients.push(peerIdFromString(peer));
            }
            await this.pubsub.publish(this.messageLog.topic, data).then((result) => {
                this.log("published %s to %d recipients %O", id, result.recipients.length, result.recipients);
                recipients.push(...result.recipients);
            }, (err) => this.log.error("failed to publish event: %O", err));
            return recipients;
        }
    }
    async handleUpdate(propagationSource, { heads }) {
        this.log("handling update: %o", heads.map(decodeId));
        const result = await this.messageLog.tree.read((txn) => {
            const missingParents = new Set();
            for (const key of heads) {
                const leaf = txn.getNode(0, key);
                if (leaf === null) {
                    missingParents.add(decodeId(key));
                }
            }
            if (missingParents.size === 0) {
                return TopicValidatorResult.Accept;
            }
            else {
                return TopicValidatorResult.Ignore;
            }
        });
        // Need to sync
        if (result === TopicValidatorResult.Ignore) {
            this.scheduleSync(propagationSource);
        }
        return result;
    }
    scheduleSync(peerId) {
        const id = peerId.toString();
        if (this.syncQueuePeers.has(id)) {
            this.log("already queued sync with %p", peerId);
            return;
        }
        this.log("scheduling sync with %p", peerId);
        this.syncQueuePeers.add(id);
        this.syncQueue
            .add(() => this.sync(peerId))
            .catch((err) => this.log.error("sync failed: %O", err))
            .finally(() => this.syncQueuePeers.delete(id));
    }
    async sync(peerId) {
        this.log("initiating sync with %p", peerId);
        const connection = this.libp2p.getConnections(peerId).find((connection) => connection.transient === false);
        if (connection === undefined) {
            this.log("no longer connected to %p", peerId);
            return;
        }
        await this.retry("sync", () => __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_sync).call(this, connection), {
            interval: SYNC_RETRY_INTERVAL,
            limit: SYNC_RETRY_LIMIT,
        });
    }
    async retry(name, callback, options) {
        for (let n = 0; n < options.limit; n++) {
            try {
                await callback();
                return;
            }
            catch (err) {
                this.log.error("%s failed: %O", name, err);
                if (this.controller.signal.aborted) {
                    break;
                }
                else {
                    const interval = Math.floor(Math.random() * options.interval);
                    this.log("waiting %dms before trying %s again (%d/%d)", interval, name, n + 1, options.limit);
                    await this.wait(interval);
                    continue;
                }
            }
        }
        throw new Error("exceeded sync retry limit");
    }
    async wait(interval) {
        await wait(interval, { signal: this.controller.signal });
    }
}
_GossipLogService_pushTopologyId = new WeakMap(), _GossipLogService_started = new WeakMap(), _GossipLogService_pushStreams = new WeakMap(), _GossipLogService_instances = new WeakSet(), _GossipLogService_sync = async function _GossipLogService_sync(connection) {
    if (connection.status !== "open") {
        throw new Error("connection closed");
    }
    const peerId = connection.remotePeer;
    const protocolSelectSignal = AbortSignal.timeout(DEFAULT_PROTOCOL_SELECT_TIMEOUT);
    const stream = await connection
        .newStream(this.syncProtocol, { negotiateFully: NEGOTIATE_FULLY, signal: protocolSelectSignal })
        .catch((err) => {
        this.log.error("failed to open outgoing sync stream: %O", err);
        throw err;
    });
    this.log("opened outgoing sync stream %s to peer %p", stream.id, peerId);
    const timeoutController = new DelayableController(SYNC_TIMEOUT);
    const signal = anySignal([this.controller.signal, timeoutController.signal]);
    signal.addEventListener("abort", (err) => {
        if (stream.status === "open") {
            stream.abort(new SyncTimeoutError());
        }
    });
    this.log("starting sync with peer %p", peerId);
    const client = new Client(stream);
    let messageCount = 0;
    try {
        await this.messageLog.sync(client, async (signedMessage) => {
            await this.messageLog.insert(signedMessage, { publish: false });
            timeoutController.delay();
            messageCount++;
        }, { peerId: peerId.toString() });
    }
    finally {
        timeoutController.clear();
        signal.clear();
        client.end();
        this.log("closed outgoing sync stream %s to peer %p", stream.id, peerId);
    }
    if (messageCount !== 0) {
        const [_, heads] = await this.messageLog.getClock();
        const data = Event.encode({ update: { heads: heads.map(encodeId) } });
        for (const peer of __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").keys()) {
            if (peer === peerId.toString())
                continue;
            __classPrivateFieldGet(this, _GossipLogService_instances, "m", _GossipLogService_push).call(this, peer, data);
        }
    }
}, _GossipLogService_push = function _GossipLogService_push(peer, data) {
    this.log("pushing data to %s", peer);
    const { stream, source } = __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").get(peer) ?? {};
    if (stream === undefined || source === undefined) {
        return;
    }
    if (stream.status !== "open") {
        __classPrivateFieldGet(this, _GossipLogService_pushStreams, "f").delete(peer);
        return;
    }
    source.push(data);
};
