import { PeerId, Startable, TypedEventEmitter, PubSub, PeerStore, TypedEventTarget, Libp2pEvents } from "@libp2p/interface";
import { Registrar, ConnectionManager } from "@libp2p/interface-internal";
import type { Signature, Message, Signer } from "@canvas-js/interfaces";
import { AbstractGossipLog, GossipLogEvents } from "../AbstractGossipLog.js";
export declare const getSyncProtocol: (topic: string) => string;
export declare const getPushProtocol: (topic: string) => string;
export type GossipLogServiceComponents = {
    peerId: PeerId;
    events: TypedEventTarget<Libp2pEvents>;
    registrar: Registrar;
    peerStore: PeerStore;
    connectionManager: ConnectionManager;
    pubsub?: PubSub;
};
export interface GossipLogServiceInit {
    maxInboundStreams?: number;
    maxOutboundStreams?: number;
}
export declare class GossipLogService<Payload = unknown> extends TypedEventEmitter<GossipLogEvents<unknown>> implements Startable {
    #private;
    private readonly components;
    readonly messageLog: AbstractGossipLog<Payload>;
    private static extractGossipSub;
    readonly syncProtocol: string;
    readonly pushProtocol: string;
    private readonly log;
    private readonly maxInboundStreams;
    private readonly maxOutboundStreams;
    private readonly syncQueue;
    private readonly syncQueuePeers;
    private readonly controller;
    private readonly pushTopology;
    private readonly litePeers;
    constructor(components: GossipLogServiceComponents, messageLog: AbstractGossipLog<Payload>, init: GossipLogServiceInit);
    isStarted(): boolean;
    start(): Promise<void>;
    private handleConnectionOpen;
    private handleConnectionClose;
    afterStart(): Promise<void>;
    beforeStop(): Promise<void>;
    stop(): Promise<void>;
    afterStop(): void;
    private forwardEvent;
    append<T extends Payload = Payload>(payload: T, options?: {
        signer?: Signer<Payload>;
    }): Promise<{
        id: string;
        signature: Signature;
        message: Message<T>;
        recipients: Promise<PeerId[]>;
    }>;
    insert(signature: Signature, message: Message<Payload>, options?: {
        peerId?: string;
    }): Promise<{
        id: string;
        recipients: Promise<PeerId[]>;
    }>;
    private publish;
    private handleMessage;
    private handleUpdate;
    private getPushSink;
    private handleIncomingPush;
    private handleIncomingSync;
    private scheduleSync;
    private sync;
    private retry;
    private wait;
}
export declare const gossiplog: <Payload>(messageLog: AbstractGossipLog<Payload>, init?: GossipLogServiceInit) => (components: GossipLogServiceComponents) => GossipLogService<Payload>;
