import { anySignal } from "any-signal";
import * as cbor from "@ipld/dag-cbor";
export const MISSING_PARENT = "MISSING_PARENT";
export class SyncDeadlockError extends Error {
}
export class SyncTimeoutError extends Error {
}
export class SyncResourceError extends Error {
}
export const cborNull = cbor.encode(null);
// eslint-disable-next-line no-useless-escape
export const topicPattern = /^[a-zA-Z0-9\.\-]+$/;
/** Logarithmic clock decay */
export function* getAncestorClocks(clock) {
    let i = 0;
    while (true) {
        const ancestor = clock - (1 << i++);
        if (ancestor > 0) {
            yield ancestor;
        }
        else {
            break;
        }
    }
}
export async function wait(interval, options) {
    if (options.signal.aborted) {
        return;
    }
    const signal = anySignal([AbortSignal.timeout(interval), options.signal]);
    await new Promise((resolve) => {
        signal.addEventListener("abort", () => resolve());
    }).finally(() => signal.clear());
}
